"""Module for dealing with time and time conventions

Some convention for this module

1. Only decorate direct spiceypy wrapper functions with the ensure_spice decorator. They should directly call a
spiceypy function.

2. All spiceypy wrapper functions should read as <spiceypyfunc>_wrapper. We really only use these to allow array
inputs for spiceypy functions that aren't already vectorized in C and to wrap them in ensure_spice.

3. All functions should have robust type-hinting.
"""
# Standard
import re
from datetime import datetime, timedelta
from typing import Union, Collection
from zoneinfo import ZoneInfo
# Installed
import numpy as np
import spiceypy as spice
# Local
from libera_utils.config import config
from libera_utils.spice_utils import ensure_spice

ISOT_REGEX = re.compile(r"^(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})"
                        r"[T|t]"
                        r"(?P<hour>[0-9]{2}):(?P<minute>[0-9]{2}):(?P<second>[0-9]{2})"
                        r"(?:\.(?P<fractional_second>[0-9]*))?$")

PRINTABLE_TS_REGEX = re.compile(r"^(?P<year>[0-9]{4})(?P<month>[0-9]{2})(?P<day>[0-9]{2})"
                                r"[T|t]"
                                r"(?P<hour>[0-9]{2})(?P<minute>[0-9]{2})(?P<second>[0-9]{2})$")

PRINTABLE_TS_FORMAT = "%Y%m%dt%H%M%S"

NUMERIC_DOY_TS_FORMAT = "%y%j%H%M%S"


def et_2_timestamp(et: Union[float, Collection[float], np.ndarray],
                   fmt: str = '%Y%m%dt%H%M%S.%f') -> Union[str, Collection[str]]:
    """
    Convert ephemeris time to a custom formatted timestamp (default is lowercase version of ISO).

    Parameters
    ----------
    et: Union[float, Collection[float], np.ndarray]
        Ephemeris Time to be converted.
    fmt: str, optional
        Format string as defined by the datetime.strftime() function.

    Returns
    -------
    : Union[str, Collection[str]]
        Formatted timestamps
    """
    datetime_objs = et_2_datetime(et)

    if isinstance(datetime_objs, Collection):
        time_out = np.array([t.strftime(fmt) for t in datetime_objs])
    else:
        time_out = datetime_objs.strftime(fmt)

    return time_out


def et_2_datetime(et: Union[float, Collection[float], np.ndarray]) -> Union[datetime, np.ndarray]:
    """
    Convert ephemeris time to a python datetime object by first converting it to a UTC timestamp.

    Parameters
    ----------
    et: Union[float, Collection[float], np.ndarray]
        Ephemeris times to be converted.

    Returns
    -------
    : Union[datetime, np.ndarray]
        Object representation of ephemeris times.
    """
    isoc_fmt = '%Y-%m-%dT%H:%M:%S.%f'
    isoc_prec = 6

    isoc_timestamp = et2utc_wrapper(et, 'ISOC', isoc_prec)
    if isinstance(et, Collection):
        return np.array([datetime.strptime(s, isoc_fmt) for s in isoc_timestamp])

    return datetime.strptime(isoc_timestamp, isoc_fmt)


@ensure_spice(time_kernels_only=True)
def et2utc_wrapper(et: Union[float, Collection[float], np.ndarray], fmt: str, prec: int) -> Union[str, Collection[str]]:
    """
    Convert ephemeris times to UTC ISO strings.
    https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/et2utc_c.html
    Decorated wrapper for spiceypy.et2utc that will automatically furnish the latest metakernel and retry
    if the first call raises an exception.

    Parameters
    ----------
    et: Union[float, Collection[float], np.ndarray]
        The ephemeris time value to be converted to UTC.
    fmt: str
        Format string defines the format of the output time string. See CSPICE docs.
    prec: int
        Number of digits of precision for fractional seconds.

    Returns
    -------
    : Union[np.ndarray, str]
        UTC time string(s)
    """
    return spice.et2utc(et, fmt, prec)


@ensure_spice(time_kernels_only=True)
def utc2et_wrapper(iso_str: Union[str, Collection[str]]) -> Union[float, np.ndarray]:
    """
    Convert UTC ISO strings to ephemeris times.
    https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/utc2et_c.html
    Decorated wrapper for spiceypy.utc2et that will automatically furnish the latest metakernel and retry
    if the first call raises an exception.

    Parameters
    ----------
    iso_str: Union[str, Collection[str]]
        The UTC to convert to ephemeris time

    Returns
    -------
    : Union[float, np.ndarray]
        Ephemeris time
    """
    if isinstance(iso_str, str):
        return spice.utc2et(iso_str)

    return np.array([spice.utc2et(s) for s in iso_str])


@ensure_spice(time_kernels_only=True)
def scs2e_wrapper(sclk_str: Union[str, Collection[str]]) -> Union[float, np.ndarray]:
    """
    Convert SCLK strings to ephemeris time.
    https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/scs2e_c.html
    Decorated wrapper for spiceypy.scs2e that will automatically furnish the latest metakernel and retry
    if the first call raises an exception.

    Parameters
    ----------
    sclk_str: Union[str, Collection[str]]
        Spacecraft clock string

    Returns
    -------
    : Union[float, np.ndarray]
        Ephemeris time
    """
    sc_id = config.get("JPSS_SC_ID")
    if isinstance(sclk_str, str):
        return spice.scs2e(sc_id, sclk_str)

    return np.array([spice.scs2e(sc_id, s) for s in sclk_str])


@ensure_spice(time_kernels_only=True)
def sce2s_wrapper(et: Union[float, Collection[float], np.ndarray]) -> Union[str, np.ndarray]:
    """
    Convert ephemeris times to SCLK string
    https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/sce2s_c.html
    Decorated wrapper for spiceypy.sce2s that will automatically furnish the latest metakernel and retry
    if the first call raises an exception.

    Parameters
    ----------
    et: Union[float, Collection[float], np.ndarray]
        Ephemeris time

    Returns
    -------
    : Union[str, Collection[str]]
        SCLK string
    """
    sc_id = config.get("JPSS_SC_ID")
    if isinstance(et, Collection):
        return np.array([spice.sce2s(sc_id, t) for t in et])

    return spice.sce2s(sc_id, et)


def convert_cds_integer_to_datetime(satellite_time: int):
    """Helper function to convert a satellite time given as an CCSDS Day Segmented Time Code (CDS) form as 8 byte
    integer to a timezone aware datetime object

    Parameters
    ----------
    satellite_time : int
        A 64-bit unsigned integer that represents CDS time

    Returns
    -------
    cds_time : datetime
     """
    byte_data = satellite_time.to_bytes(8, 'big')
    int_days = int.from_bytes([byte_data[0], byte_data[1]], byteorder="big")
    int_millisec = int.from_bytes([byte_data[2], byte_data[3],
                                    byte_data[4], byte_data[5]],
                                   byteorder="big")
    int_microsec = int.from_bytes([byte_data[6], byte_data[7]], byteorder="big")

    reference_date = datetime(1958, 1, 1, 0, 0, 0, 0, ZoneInfo("UTC"))
    cds_time = (reference_date +
                timedelta(days=int_days) +
                timedelta(milliseconds=int_millisec) +
                timedelta(microseconds=int_microsec))

    # TODO: Check with EDOS on this time conversion. The commented out below gives approximately a 70 second difference
    # TODO: to the method above.
    # TODO: satellite_time_string = f"{int_days}:{int_millisec}:{int_microsec}"
    # TODO: non_tz_datetime = et_2_datetime(scs2e_wrapper(satellite_time_string))
    # TODO: cds_time = timezone("UTC").localize(non_tz_datetime)

    return cds_time



"""Module for manifest file handling"""
# Standard
import datetime
import logging
from pathlib import Path
from bitstring import ConstBitStream
# Installed
from cloudpathlib import S3Path, AnyPath
# Local
from libera_utils.io.smart_open import smart_open
import libera_utils.db.models as libera_db_models
from libera_utils.time import convert_cds_integer_to_datetime

logger = logging.getLogger(__name__)


class EDOSGeneratedFillDataFromAPID:
    """
    Object representation of the information pertaining to which data in a Production Data Set (PDS) are generated and
    filled by EDOS. This corresponds to a database table and connects to a Construction Record (CR). This object is
    created as part of reading in a CR and requires an open bitstream to read from.
    """
    def __init__(self, cr_bitstream: ConstBitStream):
        self.ssc_with_generated_data = cr_bitstream.read("uint:32")
        self.filled_byte_offset = cr_bitstream.read("uint:64")
        self.index_to_fill_octet = cr_bitstream.read("uint:32")

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        return libera_db_models.CrApidEdosGeneratedFillData(
            ssc_with_generated_data=self.ssc_with_generated_data,
            filled_byte_offset=self.filled_byte_offset,
            index_to_fill_octet=self.index_to_fill_octet
        )


class SSCLengthDiscrepancy:
    """
    Object representation of the information of the length discrepancy in an SSC. This corresponds to a database table
    and connects to a Construction Record (CR). This object is created as part of reading in a CR and thus requires an
    open bitstream to read from.
    """
    def __init__(self, cr_bitstream: ConstBitStream):
        self.ssc_length_discrepancy = cr_bitstream.read("uint:32")

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        return libera_db_models.CrApidSscLenDiscrepancies(
            ssc_length_discrepancy=self.ssc_length_discrepancy
        )


class SCSStartStopTimes:
    """
    Object representation of the information of Spacecraft Session (SCS) start and stop times of data. This
    corresponds to a database table and connects to a Construction Record (CR). This object is created as part of
    reading in a CR and thus requires an open bitstream to read from.
    """
    def __init__(self, cr_bitstream: ConstBitStream):
        self.scs_start_time_sc_time = cr_bitstream.read("uint:64")
        self.scs_start_time_utc = convert_cds_integer_to_datetime(self.scs_start_time_sc_time)
        self.scs_stop_time_sc_time = cr_bitstream.read("uint:64")
        self.scs_stop_time_utc = convert_cds_integer_to_datetime(self.scs_stop_time_sc_time)

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        return libera_db_models.CrScsStartStopTimes(
            scs_start_sc_time=self.scs_start_time_sc_time,
            scs_stop_sc_time=self.scs_stop_time_sc_time,
            scs_start_utc_time=self.scs_start_time_utc,
            scs_stop_utc_time=self.scs_stop_time_utc
        )


class APIDFromPDSFromConstructionRecord:
    """
    Object representation of the information of Application IDs (APID) from within a Production Data Set (PDS). This
    corresponds to a database table and connects to a Construction Record (CR). This object is created as part of
    reading in a CR and thus requires an open bitstream to read from.
    """
    def __init__(self, cr_bitstream: ConstBitStream):
        # Read unused data
        cr_bitstream.read("uint:8")
        self.scid_apid = cr_bitstream.read("uint:24")
        self.apid_first_packet_sc_time = cr_bitstream.read("uint:64")
        self.apid_first_packet_utc = convert_cds_integer_to_datetime(self.apid_first_packet_sc_time)
        self.apid_last_packet_sc_time = cr_bitstream.read("uint:64")
        self.apid_last_packet_utc = convert_cds_integer_to_datetime(self.apid_last_packet_sc_time)
        # Read unused data
        cr_bitstream.read("uint:32")

    @property
    def scid(self):
        """Property that contains the SCID alone"""
        bytes_scid = self.scid_apid.to_bytes(3, 'big')
        scid_read = ConstBitStream(bytes_scid)
        return scid_read.read("uint:8")

    @property
    def apid(self):
        """Property that contains the APID alone"""
        bytes_scid = self.scid_apid.to_bytes(3, 'big')
        scid_read = ConstBitStream(bytes_scid)
        # Read unused data
        scid_read.read("uint:13")
        return scid_read.read("uint:11")

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        return libera_db_models.PdsFileApid(
                scid_apid=self.scid_apid,
                first_packet_sc_time=self.apid_first_packet_sc_time,
                last_packet_sc_time=self.apid_last_packet_sc_time,
                first_packet_utc_time=self.apid_first_packet_utc,
                last_packet_utc_time=self.apid_last_packet_utc
            )


class PDSRecord:
    """
    Object representation of a Production Data Set (PDS). This object is created as part of
    reading in a PDS filename and storing that filename and ingest time in a database.
    """
    def __init__(self, pds_filename: str):
        self.pds_filename = pds_filename

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        return libera_db_models.PdsFile(
            file_name=self.pds_filename,
            ingested=datetime.datetime.utcnow()
        )


class PDSFileFromConstructionRecord:
    """
    Object representation of the information directly related to a Production Data Set (PDS). This corresponds to a
    database table and connects to a Construction Record (CR). This object is created as part of reading in a CR and
    requires an open bitstream to read from.
    """
    def __init__(self, cr_bitstream: ConstBitStream):
        self.pds_filename = (cr_bitstream.read("bytes:40")).decode()
        # Read unused data
        cr_bitstream.read("uint:24")

        # This is quoted in 25-4 as a "one-up" counter with values of 1 to 3. However, there is a situation when the
        # value can be 0, and then there is one entry with complete data as 0's throughout. To account for this take
        # the maximum of the value and 1 to ensure if a 0 is there at least one full entry of 0's is read.
        self.apid_count_this_file = max(cr_bitstream.read("uint:8"), 1)
        self.apid_this_file = []
        for _ in range(self.apid_count_this_file):
            self.apid_this_file.append(APIDFromPDSFromConstructionRecord(cr_bitstream))

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        apids_list = []
        for i in range(self.apid_count_this_file):
            apids_list.append(self.apid_this_file[i].to_orm())
        return libera_db_models.PdsFile(
            file_name=self.pds_filename,
            apids=apids_list
        )


class SSCGapInformationFromConstructionRecord:
    """
    Object representation of the information of Spacecraft Contact Sessions (SCS). This corresponds to a database table
    and connects to a Construction Record (CR). This object is created as part of reading in a CR and thus requires an
    open bitstream to read from.
    """
    def __init__(self, cr_bitstream: ConstBitStream):
        self.apid_gap_first_missing_ssc_packet = cr_bitstream.read("uint:32")
        self.apid_gap_byte_offset = cr_bitstream.read("uint:64")
        self.apid_num_ssc_packets_missed = cr_bitstream.read("uint:32")

        # These are not labeled in the ICD document and so this is a guess based on other patterns in the ICD
        sc_packet_before_time = cr_bitstream.read("uint:64")
        sc_packet_after_time = cr_bitstream.read("uint:64")
        self.apid_preceding_packet_sc_time = sc_packet_before_time
        self.apid_following_packet_sc_time = sc_packet_after_time
        self.apid_preceding_packet_utc = convert_cds_integer_to_datetime(sc_packet_before_time)
        self.apid_following_packet_utc = convert_cds_integer_to_datetime(sc_packet_after_time)

        self.apid_preceding_packet_esh_time = cr_bitstream.read("uint:64")
        self.apid_following_packet_esh_time = cr_bitstream.read("uint:64")

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        return libera_db_models.CrApidSscGap(
            first_missing_ssc=self.apid_gap_first_missing_ssc_packet,
            gap_byte_offset=self.apid_gap_byte_offset,
            n_missing_sscs=self.apid_num_ssc_packets_missed,
            preceding_packet_sc_time=self.apid_preceding_packet_sc_time,
            following_packet_sc_time=self.apid_following_packet_sc_time,
            preceding_packet_utc_time=self.apid_preceding_packet_utc,
            following_packet_utc_time=self.apid_following_packet_utc,
            preceding_packet_esh_time=self.apid_preceding_packet_esh_time,
            following_packet_esh_time=self.apid_following_packet_esh_time
        )


class VCIDFromConstructionRecord:
    """
    Object representation of the information of Virtual Channel ID (VCID). This corresponds to a database table and
    connects to a Construction Record (CR). This object is created as part of reading in a CR and thus requires an
    open bitstream to read from.
    """
    def __init__(self, cr_bitstream: ConstBitStream):
        self.vcid_scid = cr_bitstream.read("uint:16")

    @property
    def scid(self):
        """Property that contains the SCID alone"""
        bytes_vcdu = self.vcid_scid.to_bytes(2, 'big')
        vcdu_read = ConstBitStream(bytes_vcdu)
        # Read unused data
        vcdu_read.read("uint:2")
        return vcdu_read.read("uint:8")

    @property
    def vcid(self):
        """Property that contains the APID alone"""
        bytes_vcdu = self.vcid_scid.to_bytes(2, 'big')
        vcdu_read = ConstBitStream(bytes_vcdu)
        # Read unused data
        vcdu_read.read("uint:10")
        return vcdu_read.read("uint:6")

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        return libera_db_models.CrApidVcid(
            scid_vcid=self.vcid_scid
        )


class APIDFromConstructionRecord:
    """
    Object representation of the information of Application IDs (APID) from within a Construction Record (CR). This
    corresponds to a database table and connects to a CR. This object is created as part of reading in a CR and thus
    requires an open bitstream to read from.
    """
    def __init__(self, cr_bitstream: ConstBitStream):
        # Read unused data
        cr_bitstream.read("uint:8")
        self.apid_scid = cr_bitstream.read("uint:24")

        self.apid_byte_offset = cr_bitstream.read("uint:64")
        # Read unused data
        cr_bitstream.read("uint:24")

        # For this APID, identify the Virtual Channel Identification (VCID(s))
        self.apid_vcid_count = cr_bitstream.read("uint:8")
        self.vcids_list = []
        for _ in range(self.apid_vcid_count):
            # Read unused data
            cr_bitstream.read("uint:16")
            self.vcids_list.append(VCIDFromConstructionRecord(cr_bitstream))

        # List missing packets SSCs for the PDS
        self.apid_ssc_gap_count = cr_bitstream.read("uint:32")
        self.apid_ssc_gaps_list = []
        for _ in range(self.apid_ssc_gap_count):
            self.apid_ssc_gaps_list.append(SSCGapInformationFromConstructionRecord(cr_bitstream))

        # For this APID, list packets containing EDOS generated fill data
        self.edos_generated_fill_data_count = cr_bitstream.read("uint:32")
        self.edos_generated_fill_data_list = []
        for _ in range(self.edos_generated_fill_data_count):
            self.edos_generated_fill_data_list.append(EDOSGeneratedFillDataFromAPID(cr_bitstream))

        self.edos_generated_octet_count = cr_bitstream.read("uint:64")
        # For the packets with length discrepancy
        self.ssc_length_discrepancy_count = cr_bitstream.read("uint:32")
        self.ssc_length_discrepancy_list = []
        for _ in range(self.ssc_length_discrepancy_count):
            self.ssc_length_discrepancy_list.append(SSCLengthDiscrepancy(cr_bitstream))

        self.first_packet_sc_time = cr_bitstream.read("uint:64")
        self.last_packet_sc_time = cr_bitstream.read("uint:64")
        self.first_packet_esh_time = cr_bitstream.read("uint:64")
        self.last_packet_esh_time = cr_bitstream.read("uint:64")

        self.first_packet_time_utc = convert_cds_integer_to_datetime(self.first_packet_sc_time)
        self.last_packet_time_utc = convert_cds_integer_to_datetime(self.last_packet_sc_time)

        self.vcdu_error_packet_count = cr_bitstream.read("uint:32")

        # This is not well labeled in the ICD (24-17)
        self.count_in_the_data_set = cr_bitstream.read("uint:32")

        self.apid_size_octets = cr_bitstream.read("uint:64")
        # Read unused data
        cr_bitstream.read("uint:64")

    @property
    def scid(self):
        """Property that contains the SCID alone"""
        bytes_scid = self.apid_scid.to_bytes(3, 'big')
        scid_read = ConstBitStream(bytes_scid)
        return scid_read.read("uint:8")

    @property
    def apid(self):
        """Property that contains the APID alone"""
        bytes_scid = self.apid_scid.to_bytes(3, 'big')
        scid_read = ConstBitStream(bytes_scid)
        # Read unused data
        scid_read.read("uint:13")
        return scid_read.read("uint:11")

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        vcids = []
        for i in range(self.apid_vcid_count):
            vcids.append(self.vcids_list[i].to_orm())
        ssc_gaps = []
        for i in range(self.apid_ssc_gap_count):
            ssc_gaps.append(self.apid_ssc_gaps_list[i].to_orm())
        edos_fill_data = []
        for i in range(self.edos_generated_fill_data_count):
            edos_fill_data.append(self.edos_generated_fill_data_list[i].to_orm())
        ssc_length_discrep = []
        for i in range(self.ssc_length_discrepancy_count):
            ssc_length_discrep.append(self.ssc_length_discrepancy_list[i].to_orm())

        return libera_db_models.CrApid(
            scid_apid=self.apid_scid,
            byte_offset=self.apid_byte_offset,
            n_vcids=self.apid_vcid_count,
            vcids=vcids,
            n_ssc_gaps=self.apid_ssc_gap_count,
            ssc_gaps=ssc_gaps,
            n_edos_generated_fill_data=self.edos_generated_fill_data_count,
            edos_fill_data=edos_fill_data,
            count_edos_generated_octets=self.edos_generated_octet_count,
            n_length_discrepancy_packets=self.ssc_length_discrepancy_count,
            ssc_length_discrepancies=ssc_length_discrep,
            first_packet_sc_time=self.first_packet_sc_time,
            last_packet_sc_time=self.last_packet_sc_time,
            esh_first_packet_time=self.first_packet_esh_time,
            esh_last_packet_time=self.last_packet_esh_time,
            first_packet_utc_time=self.first_packet_time_utc,
            last_packet_utc_time=self.last_packet_time_utc,
            n_vcdu_corrected_packets=self.vcdu_error_packet_count,
            n_in_the_data_set=self.count_in_the_data_set,
            n_octect_in_apid=self.apid_size_octets
        )


class ConstructionRecordError(Exception):
    """Generic exception related to construction record file handling"""
    pass


class ConstructionRecord:
    """
    Object representation of a JPSS Construction Record (CR) including objects for all the other classes
    in this file to be stored in a database.
    """
    @classmethod
    def from_file(cls, filepath: str or Path or S3Path):
        """Read a construction record file and return a ConstructionRecord object (factory method).

            Parameters
            ----------
            filepath : str or Path or S3Path
                Location of construction record file to read.

            Returns
            -------
            : ConstructionRecord
        """
        with smart_open(filepath) as const_record_file:
            cr_bitstream = ConstBitStream(const_record_file)
            return cls(filepath, cr_bitstream)

    def __init__(self, filepath: str or Path or S3Path, cr_bitstream: ConstBitStream):
        path_object = AnyPath(filepath)
        # Any Posix Path will have the member 'name' so disable pylint on this line
        self.file_name = path_object.name  # pylint: disable=no-member
        self.edos_version = cr_bitstream.read("uint:16")

        # Construction Record type 1 is for PDS
        self.construction_record_type = cr_bitstream.read("uint:8")
        # Read unused data
        cr_bitstream.read("uint:8")
        self.cr_id = (cr_bitstream.read("bytes:36")).decode()
        # Read unused data
        cr_bitstream.read("uint:7")
        self.test_flag = cr_bitstream.read("bool")
        # Read unused data
        cr_bitstream.read("uint:8")
        cr_bitstream.read("uint:64")

        self.scs_num_start_stop_times = cr_bitstream.read("uint:16")
        self.scs_start_stop_times_list = []
        for _ in range(self.scs_num_start_stop_times):
            self.scs_start_stop_times_list.append(SCSStartStopTimes(cr_bitstream))

        self.pds_num_bytes_fill_data = cr_bitstream.read("uint:64")
        self.pds_packet_length_mismatch_count = cr_bitstream.read("uint:32")
        self.pds_first_packet_sc_time = cr_bitstream.read("uint:64")
        self.pds_first_packet_utc_time = convert_cds_integer_to_datetime(self.pds_first_packet_sc_time)
        self.pds_last_packet_sc_time = cr_bitstream.read("uint:64")
        self.pds_last_packet_utc_time = convert_cds_integer_to_datetime(self.pds_last_packet_sc_time)
        self.pds_first_packet_esh_time = cr_bitstream.read("uint:64")
        self.pds_last_packet_esh_time = cr_bitstream.read("uint:64")
        self.pds_rs_corrected_count = cr_bitstream.read("uint:32")
        self.pds_packet_count = cr_bitstream.read("uint:32")
        self.pds_size = cr_bitstream.read("uint:64")
        self.pds_discontinuities_count = cr_bitstream.read("uint:32")
        self.pds_completion_time_bytes = cr_bitstream.read("uint:64")
        # Read unused data
        cr_bitstream.read("uint:56")

        # For the PDS, identify the APIDs and their associated information.
        self.apid_count = cr_bitstream.read("uint:8")
        self.apid_data_list = []
        for _ in range(self.apid_count):
            self.apid_data_list.append(APIDFromConstructionRecord(cr_bitstream))

        # Read unused data
        cr_bitstream.read("uint:24")
        # Identify files that store this PDS
        self.pds_file_count = cr_bitstream.read("uint:8")
        self.pds_files_list = []
        for _ in range(self.pds_file_count):
            self.pds_files_list.append(PDSFileFromConstructionRecord(cr_bitstream))

    @property
    def edos_version_major(self):
        """Property that contains the major version number of the EDOS software alone"""
        edos_bytes = self.edos_version.to_bytes(2, 'big')
        edos_read = ConstBitStream(edos_bytes)
        return edos_read.read("uint:8")

    @property
    def edos_version_release(self):
        """Property that contains the major version release number of the EDOS software alone"""
        edos_bytes = self.edos_version.to_bytes(2, 'big')
        edos_read = ConstBitStream(edos_bytes)
        edos_read.read("uint:8")
        return edos_read.read("uint:8")

    def to_orm(self):
        """Convert this class instance to a corresponding ORM object for entry into the database"""
        scs_start_stops = []
        for i in range(self.scs_num_start_stop_times):
            scs_start_stops.append(self.scs_start_stop_times_list[i].to_orm())
        apids = []
        for i in range(self.apid_count):
            apids.append(self.apid_data_list[i].to_orm())
        pds_files = []
        for i, pds_file in enumerate(self.pds_files_list):
            pds_files.append(pds_file.to_orm())
        return libera_db_models.Cr(
            file_name=self.file_name,
            edos_software_version=self.edos_version,
            construction_record_type=self.construction_record_type,
            test_flag=self.test_flag,
            n_scs_start_stops=self.scs_num_start_stop_times,
            scs_start_stop_times=scs_start_stops,
            n_bytes_fill_data=self.pds_num_bytes_fill_data,
            n_length_mismatches=self.pds_packet_length_mismatch_count,
            first_packet_sc_time=self.pds_first_packet_sc_time,
            last_packet_sc_time=self.pds_last_packet_sc_time,
            first_packet_utc_time=self.pds_first_packet_utc_time,
            last_packet_utc_time=self.pds_last_packet_utc_time,
            first_packet_esh_time=self.pds_first_packet_esh_time,
            last_packet_esh_time=self.pds_last_packet_esh_time,
            n_rs_corrections=self.pds_rs_corrected_count,
            n_packets=self.pds_packet_count,
            size_bytes=self.pds_size,
            n_ssc_discontinuities=self.pds_discontinuities_count,
            completion_time=self.pds_completion_time_bytes,
            n_apids=self.apid_count,
            apids=apids,
            n_pds_files=self.pds_file_count,
            pds_files=pds_files
        )



"""Quality flag definitions"""
# Standard
from operator import or_ as _or_
from functools import reduce
# Local
from libera_utils.backports.enum_3_11 import Flag, EnumMeta, Enum, STRICT


# pylint: disable-all
class FrozenFlagMeta(EnumMeta):
    """
    Metaclass that freezes an enum entirely, preventing values from being updated, added, or deleted.
    """
    def __new__(mcs, name, bases, classdict):
        classdict['__frozenenummeta_creating_class__'] = True
        flag = super().__new__(mcs, name, bases, classdict)
        del flag.__frozenenummeta_creating_class__
        return flag

    def __call__(cls, value, names=None, *, module=None, **kwargs):
        if names is None:  # simple value lookup
            return cls.__new__(cls, value)
        enum = Enum._create_(value, names, module=module, **kwargs)
        enum.__class__ = type(cls)
        return enum

    def __setattr__(cls, name, value):
        members = cls.__dict__.get('_member_map_', {})
        if hasattr(cls, '__frozenenummeta_creating_class__') or name in members:
            return super().__setattr__(name, value)
        if hasattr(cls, name):
            msg = "{!r} object attribute {!r} is read-only"
        else:
            msg = "{!r} object has no attribute {!r}"
        raise AttributeError(msg.format(cls.__name__, name))

    def __delattr__(cls, name):
        members = cls.__dict__.get('_member_map_', {})
        if hasattr(cls, '__frozenenummeta_creating_class__') or name in members:
            return super().__delattr__(name)
        if hasattr(cls, name):
            msg = "{!r} object attribute {!r} is read-only"
        else:
            msg = "{!r} object has no attribute {!r}"
        raise AttributeError(msg.format(cls.__name__, name))


class QualityFlag(Flag, boundary=STRICT):
    """
    Subclass of Flag that add a method for decomposing a flag into its individual components
    and a property to return a list of all messages associated with a quality flag
    """

    def decompose(self):
        """
        Return the set of all set flags that form a subset of the queried flag value. Note that this is not the
        minimum set of quality flags but rather a full set of all flags such that when they are ORed together, they
        produce self.value

        :return: members, not_covered
            `members` is a list of flag values that are subsets of `value`
            `not_covered` is zero if the OR of members recreates `value`. Non-zero otherwise if bits are set in `value`
                that do not exist as named values in cls.
        """
        value = self.value
        not_covered = value
        flags_to_check = [  # Creates the "basis" for the quality flag
            (m, v)
            for v, m in list(self.__class__._value2member_map_.items())
            if m.name in (x.name for x in self)
        ]
        members = []
        for member, member_value in flags_to_check:
            if member_value and member_value & value == member_value:
                members.append(member)
                not_covered &= ~member_value
        if not members and value in self.__class__._value2member_map_:
            members.append(self.__class__._value2member_map_[value])
        members.sort(key=lambda m: m._value_, reverse=True)
        return members, not_covered

    @property
    def summary(self):
        """Summarize quality flag value"""
        members, not_covered = self.decompose()
        print(members)
        if not_covered:
            raise ValueError(f"{self.__name__} has value {self.value} but that value cannot be created by elements "
                             f"of {self.__class__}. This should never happen unless a quality flag was declared "
                             f"without using the FrozenFlagMeta metaclass.")

        try:
            return int(self.value), [m.value.message for m in members]
        except Exception as err:
            raise AttributeError(
                "Tried to summarize a quality flag but its values don't appear to have messages.") from err


class FlagBit(int):
    """Subclass of int to capture both an integer value and an accompanying message"""
    def __new__(cls, *args, message=None, **kwargs):
        obj = super().__new__(cls, *args, **kwargs)
        obj.message = message
        return obj

    def __str__(self):
        return f"{super().__str__()}: {self.message}"


def with_all_none(f):
    """Add NONE and ALL psuedo-members to f"""
    f._member_map_['NONE'] = f(FlagBit(0, message="No flags set."))
    f._member_map_['ALL'] = f(reduce(_or_, f))
    return f



"""Module for performing geolocation tasks"""
# Standard
import logging
# Installed
import numpy as np
import spiceypy as spice
# Local
from libera_utils import spice_utils

logger = logging.getLogger(__name__)


vec_pxform = np.vectorize(spice.pxform,
                          excluded=['fromstr', 'tostr'],
                          signature='(),(),()->(3,3)',
                          otypes=[np.float64])
"""
Vectorized form of spice.pxform function.

"Return the matrix that transforms position vectors from one specified frame to another at a specified epoch."
https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/pxform_c.html

======== ==== ===============================================
VARIABLE  I/O  DESCRIPTION
======== ==== ===============================================
from       I   Name of the frame to transform from.
to         I   Name of the frame to transform to.
et         I   Epoch of the rotation matrix.
rotate     O   A rotation matrix.
======== ==== ===============================================

Note: For time-independent reference frames transformations, the ET value passed doesn't matter,
as long as it is a valid ephemeris time.
"""


vec_subpnt = np.vectorize(spice.subpnt,
                          excluded=['method', 'target', 'fixref', 'abcorr', 'obsrvr'],
                          signature='(),(),(),(),(),()->(3),(),(3)',
                          otypes=[np.float64, np.float64, np.float64])
"""
Vectorized form of spice.subpnt function

"Compute the rectangular coordinates of the sub-observer point on a target body at a specified epoch, optionally
corrected for light time and stellar aberration."
https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/subpnt_c.html

======== ==== ===============================================
Variable  I/O  Description
======== ==== ===============================================
method     I   Computation method.
target     I   Name of target body.
et         I   Epoch in TDB seconds past J2000 TDB.
fixref     I   Body-fixed, body-centered target body frame.
abcorr     I   Aberration correction flag.
obsrvr     I   Name of observing body.
spoint     O   Sub-observer point on the target body.
trgepc     O   Sub-observer point epoch.
srfvec     O   Vector from observer to sub-observer point.
======== ==== ===============================================
"""


vec_subslr = np.vectorize(spice.subslr,
                          excluded=['method', 'target', 'fixref', 'abcorr', 'obsrvr'],
                          signature='(),(),(),(),(),()->(3),(),(3)',
                          otypes=[np.float64, np.float64, np.float64])
"""
Vectorized form of spice.subslr function

"Compute the rectangular coordinates of the sub-solar point on a target body at a specified epoch, optionally
corrected for light time and stellar aberration."
https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/subslr_c.html

======== ==== ===============================================
Variable  I/O  Description
======== ==== ===============================================
method     I   Computation method.
target     I   Name of target body.
et         I   Epoch in ephemeris seconds past J2000 TDB.
fixref     I   Body-fixed, body-centered target body frame.
abcorr     I   Aberration correction.
obsrvr     I   Name of observing body.
spoint     O   Sub-solar point on the target body.
trgepc     O   Sub-solar point epoch.
srfvec     O   Vector from observer to sub-solar point.
======== ==== ===============================================
"""

vec_npedln = np.vectorize(spice.npedln,
                          excluded=['a', 'b', 'c'],
                          signature='(),(),(),(3),(3)->(3),()',
                          otypes=[np.float64, np.float64])
"""
Vectorized form of spice.subslr function

"Compute the rectangular coordinates of the sub-solar point on a target body at a specified epoch, optionally
corrected for light time and stellar aberration."
https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/subslr_c.html

======== ==== ===============================================
Variable  I/O  Description
======== ==== ===============================================
method     I   Computation method.
target     I   Name of target body.
et         I   Epoch in ephemeris seconds past J2000 TDB.
fixref     I   Body-fixed, body-centered target body frame.
abcorr     I   Aberration correction.
obsrvr     I   Name of observing body.
spoint     O   Sub-solar point on the target body.
trgepc     O   Sub-solar point epoch.
srfvec     O   Vector from observer to sub-solar point.
======== ==== ===============================================
"""

# TODO: Try moving the ensure_spice decorators
#  off of the higher level functions below and onto the vectorized spice functions above.


@spice_utils.ensure_spice
def get_earth_radii():
    """
    Retrieve Earth radii values from SPICE

    Returns
    -------
    : tuple
        (re, rp, flat) tuple of equatorial and polar ellipsoid radii and a flattening coefficient
    """
    # Retrieve Earth radii
    _, radii = spice.bodvrd(spice_utils.SpiceBody.EARTH.value.strid, 'RADII', 3)
    re = radii[0]  # Equatorial
    rp = radii[2]  # Polar
    flat = (re - rp) / re
    return re, rp, flat


@spice_utils.ensure_spice
def target_position(target: spice_utils.SpiceBody, et: float or np.ndarray,
                    frame: spice_utils.SpiceFrame, observer: spice_utils.SpiceBody,
                    abcorr: str = 'NONE', normalize: bool = False):
    """Calculates the position and velocity of the `target` at ephemeris time `et` relative to `observer` in
    reference frame `frame`.
    Also calculates the light travel time between `target` and `observer` at time `et`.

    Parameters
    ----------
    target : spice_utils.SpiceBody
        Target body for which to calculate position and velocity relative to observer
    et : float or numpy array
        Ephemeris time(s)
    frame : spice_utils.SpiceFrame
        Reference frame (unit vectors)
    observer : spice_utils.SpiceBody
        The observer of the target. Resulting coordinates point from observer to target.
    abcorr : str
        A scalar string that indicates the aberration corrections to apply to the database of the target body to account
        for one-way light time and stellar aberration. Default is 'NONE'.
    normalize : bool, Optional
        Return unit vectors for position and velocity (light time output is unchanged)

    Returns
    -------
    : tuple
        (x: np.ndarray, v: np.ndarray, lt: np.ndarray) or (x: float, v: float, lt: float)
        Rectangular position and velocity vectors (x, y, z), (v_x, v_y, v_z) where position
        points from the planet center of mass location at ``et`` to the aberration-corrected location of the target.
        Light time (lt) between planetary body and target.
    """
    # https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/spkezr_c.html
    position_velocity, light_time_between = spice.spkezr(target.value.strid, et,
                                                         frame.value.strid, abcorr, observer.value.strid)

    if isinstance(position_velocity, list):
        position = np.array([el[0:3] for el in position_velocity])
        velocity = np.array([el[3:6] for el in position_velocity])
        if normalize:
            position = np.array([v / np.linalg.norm(v) for v in position])
            velocity = np.array([v / np.linalg.norm(v) for v in velocity])
    else:
        position = np.array(position_velocity[0:3])
        velocity = np.array(position_velocity[3:6])
        if normalize:
            position = position / np.linalg.norm(position)
            velocity = velocity / np.linalg.norm(velocity)
    return position, velocity, light_time_between


@spice_utils.ensure_spice
def sub_observer_point(target: spice_utils.SpiceBody, et: float or np.ndarray,
                       frame: spice_utils.SpiceFrame, observer: spice_utils.SpiceBody,
                       abcorr: str = 'NONE', method: str = 'NEAR POINT/ELLIPSOID'):
    """Computes the cartesian coordinates of the sub-observer point at time `et` and the observer altitude
    above the point. Units in km.

    Parameters
    ----------
    target : spice_utils.SpiceBody
        Body on which the sub point will be calculated (usually a planetary body).
    et : float or np.ndarray
        Ephemeris time of observation.
    frame : spice_utils.SpiceFrame
        Reference frame for returned vectors.
    observer : spice_utils.SpiceBody
        The object from which to calculate the sub point (e.g. a spacecraft). A-B correction is applied based on the
        distance between observer and sub observer point.
    abcorr : str, Optional
        String delineating what kind of A-B light time correction to perform. Default is 'NONE'.
    method : str, Optional
        String specifying what kind of method to use to find the vector between the observer and the target.
        Default is `NEAR POINT/ELLIPSOID`, which uses the nearest point on the ellipsoid rather than drawing a line
        through the center of the ellipsoid.

    Returns
    -------
    : np.ndarray, float
        Cartesian point on the target body surface in the specified reference frame
        and also the euclidean distance between the observer and the sub point in order: [x, y, z], obs_alt
    """

    spoint, _, obs_tgt_vec = vec_subpnt(method, target.value.strid, et, frame.value.strid,
                                             abcorr, observer.value.strid)
    if hasattr(et, '__len__'):
        observer_alt = np.linalg.norm(obs_tgt_vec, axis=1)  # Calculate distance from obs to spoint as observer altitude
    else:
        observer_alt = np.linalg.norm(obs_tgt_vec)
    return spoint, observer_alt


@spice_utils.ensure_spice
def sub_solar_point(target: spice_utils.SpiceBody, et: float or np.ndarray,
                    frame: spice_utils.SpiceFrame, observer: spice_utils.SpiceBody,
                    abcorr='LT+S', method='NEAR POINT/ELLIPSOID'):
    """Computes the cartesian coordinates of the subsolar point at ephemeris time et.

    Parameters
    ----------
    target : spice_utils.SpiceBody
        Body on which the sub point will be calculated (usually a planetary body).
    et : float or np.ndarray
        Ephemeris time of observation.
    frame : spice_utils.SpiceFrame
        Reference frame for returned vectors.
    observer : spice_utils.SpiceBody
        The object from which to calculate the subsolar point (e.g. a spacecraft).
        A-B correction is applied based on the distance between observer and subsolar point.
    abcorr : str
        String delineating what kind of A-B lighttime correction to perform. Default is 'LT+S'.
    method : str
        String specifying what kind of method to use to find the vector between the observer and the target.
        Default is `NEAR POINT/ELLIPSOID`, which uses the nearest point on the ellipsoid rather than drawing a line
        through the center of the ellipsoid.

    Returns
    -------
    : np.ndarray, np.ndarray, np.ndarray
        Subsolar point on the ellipsoid surface in the specified reference frame, apparent epoch at that point
        (depending on specified light time correction), and vector from observer to subsolar point.
    """
    fixref = frame.value.strid
    obsrvr = observer.value.strid
    spoint, trgepc, srfvec = vec_subslr(method, target.value.strid, et, fixref, abcorr, obsrvr)

    return spoint, trgepc, srfvec


@spice_utils.ensure_spice
def frame_transform(from_frame: spice_utils.SpiceFrame, to_frame: spice_utils.SpiceFrame,
                    et: float or np.ndarray, position: np.ndarray,
                    normalize: bool = False) -> np.ndarray:
    """
    Transform a position <x, y, z> vector between reference frames, optionally normalizing the result.

    Parameters
    ----------
    from_frame : spice_utils.SpiceFrame
        Reference frame of position
    to_frame : spice_utils.SpiceFrame
        Reference frame of output
    et : np.float64 or np.ndarray with dtype np.float64
        Ephemeris time(s) corresponding to position(s).
        For time-independent transformations, this can by any valid ephemeris time.
    position : np.ndarray
        <x, y, z> vector or array of vectors in reference frame `from_frame`
    normalize : bool, Optional
        Optionally normalize the output vector

    Returns
    -------
    : np.ndarray
        3d position vector(s) in reference frame `to_frame`
    """
    if ((position.ndim == 1 and not (len(position) == 3 and isinstance(et, float))) or
            (position.ndim == 2 and not (len(position) == len(et) and len(position[0]) == 3))):
        raise ValueError("Incorrect dimensions.")

    rotate = vec_pxform(from_frame.value.strid, to_frame.value.strid, et)

    if hasattr(rotate[0][0], '__len__'):
        result = np.array([np.matmul(rotate, pos).astype(np.float64) for rotate, pos in zip(rotate, position)])
        if normalize:
            result = np.array([r / np.linalg.norm(r) for r in result])
    else:
        result = np.matmul(rotate, position).astype(np.float64)
        if normalize:
            result = result / np.linalg.norm(result)

    return result


def angle_between(v1: np.ndarray, v2: np.ndarray, degrees: bool = False):
    """
    Returns angle between vectors v1 and v2, in units of radians (default) or degrees.
    N is the number of vectors
    D is the dimension of the space

    Parameters
    ----------
    v1 : np.ndarray
        Vector(s) 1. May be shape (D,) or (N, D).
    v2 : nd.ndarray
        Vector(s) 2. May be shape (D,) or (N, D).
    degrees : bool
        Specify True to return result in degrees. Default is False (returns radians).

    Returns
    -------
    : float or np.ndarray
        Angle between v1 and v2 in radians (optionally in degrees)
    """
    if v1.ndim == 1 and v2.ndim == 1:
        u1 = v1 / np.linalg.norm(v1)
        u2 = v2 / np.linalg.norm(v2)
        theta = np.arccos(np.clip(np.dot(u1, u2), -1.0, 1.0))
    elif v1.ndim == 2 and v2.ndim == 2:
        u1 = v1 / np.array([np.linalg.norm(v1, axis=1)]).T
        u2 = v2 / np.array([np.linalg.norm(v2, axis=1)]).T
        theta = np.array([np.arccos(np.clip(np.dot(a, b), -1.0, 1.0)) for a, b in zip(u1, u2)])
    else:
        raise ValueError(f"Function {__name__} only accepts inputs of shape (3,) or (3, N). "
                         f"Got v1.shape={v1.shape}, v2.shape={v2.shape}.")

    if degrees:
        theta = np.rad2deg(theta)
    return theta


@spice_utils.ensure_spice
def cartesian_to_planetographic(cartesian_coords: np.ndarray, degrees: bool = True):
    """
    Convert cartesian coordinates in the ITRF93 frame to planetographic latitude and longitude.
    Longitude runs 0-360 such that longitude appears to increase
    as the planet rotates when viewed by an observer and latitude is calculated from a surface normal vector rather
    than a line through the planet center. See
    https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/Tutorials/pdf/individual_docs/17_frames_and_coordinate_systems.pdf
    for reference.

    Parameters
    ----------
    cartesian_coords : np.ndarray
        Rectangular coordinates in ITRF93 frame.
    degrees : bool
        Default true. If False, returns angles in radians.

    Returns
    -------
    : np.ndarray
        Each coordinate is returned as (longitude, latitude, altitude).
    """
    re, _, flat = get_earth_radii()

    # https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/recpgr_c.html
    if hasattr(cartesian_coords[0], '__len__'):
        nan_locations = np.any(np.isnan(cartesian_coords), axis=1)
        result = np.array([
            spice.recpgr(spice_utils.SpiceBody.EARTH.value.strid, p.astype(np.float64), re, flat)
            for p in cartesian_coords
        ])
        longitude = result[:, 0]
        latitude = result[:, 1]
        altitude = result[:, 2]
        longitude[nan_locations] = np.nan
        latitude[nan_locations] = np.nan
        altitude[nan_locations] = np.nan
    else:
        if np.any(np.isnan(cartesian_coords.astype(np.float64))):
            longitude, latitude, altitude = (np.nan, np.nan, np.nan)
        else:
            longitude, latitude, altitude = spice.recpgr(
                spice_utils.SpiceBody.EARTH.value.strid, cartesian_coords.astype(np.float64), re, flat)

    if degrees:
        return np.rad2deg(longitude), np.rad2deg(latitude), altitude

    return longitude, latitude, altitude


@spice_utils.ensure_spice
def surface_intercept_point(sc_location: np.ndarray, look_vector: np.ndarray,
                            look_frame: spice_utils.SpiceFrame, et: float or np.ndarray = None):
    """
    Returns rectangular coordinates of the point of interception of a look direction from the spacecraft onto
    the Earth ellipsoid. If the look vector misses the planet, then the distance returned will be non-zero and the
    point returned is the point on the look_vector ray that is closest to the ellipsoid.

    This routine assumes that the location of the spacecraft and the location of the instrument are the same
    because we don't have ephemeris data for the instrument but we _do_ have ephemeris for the spacecraft. Over
    the scale of distances involved, the offset between spacecraft and instrument (meters) should be negligible
    in affecting the near-point calculation.

    https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/cspice/npedln_c.html

    Parameters
    ----------
    sc_location : np.ndarray
        The location of the observing body (i.e. the spacecraft body) with respect to Earth
    look_vector : np.ndarray
        Look direction unit vector (e.g. an instrument look direction)
    look_frame : spice_utils.SpiceFrame
        Reference frame of `look_vector`
    et : float or np.ndarray or None, Optional
        Ephemeris time (at spacecraft at photon detection time). Only required if look_frame is not ITRF93.

    Returns
    -------
    : tuple (pnear, alt)
        Rectangular coordinates of nearest point to reference surface ellipsoid and distance
        between the line and the near point.
    """
    # Transform look unit vector(s) from instrument frame into the ITRF93 frame
    if look_frame != spice_utils.SpiceFrame.ITRF93:
        if not et:
            raise ValueError("You specified a look_frame other than ITRF93 but no ephemeris time. "
                             "ET is required to transform your look vector to ITRF93 frame.")
        look_vector_ecef = frame_transform(look_frame, spice_utils.SpiceFrame.ITRF93, et, look_vector, normalize=True)
    else:
        look_vector_ecef = look_vector

    re_earth, rp_earth, _ = get_earth_radii()

    if sc_location.ndim == 2:  # More than one sc_location and look_vector
        pnear_dist = [spice.npedln(re_earth, re_earth, rp_earth, loc, look)
                      for loc, look in zip(sc_location, look_vector_ecef)]
        # [([x, y, z], d), ...]
        pnear = np.array([p for p, d in pnear_dist])  # [[x1, y1, z1], [x2, y2, z2], ...]
        dist = np.array([d for p, d in pnear_dist])  # [d1, d2, ...]
    else:  # Just one sc_location and look_vector
        pnear, dist = spice.npedln(re_earth, re_earth, rp_earth, sc_location, look_vector_ecef)
    return pnear, dist



"""Module for file naming utilities"""
# Standard
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from importlib import metadata
import os
import re
from types import SimpleNamespace
from pathlib import Path
# Installed
from cloudpathlib import AnyPath, CloudPath
# Local
from libera_utils.time import PRINTABLE_TS_FORMAT, NUMERIC_DOY_TS_FORMAT

REVISION_TS_FORMAT = f"r{NUMERIC_DOY_TS_FORMAT}"  # Just adds an r in front

SPK_REGEX = re.compile(r"^libera_(?P<spk_object>jpss)"
                       r"_(?P<utc_start>[0-9]{8}t[0-9]{6})"
                       r"_(?P<utc_end>[0-9]{8}t[0-9]{6})"
                       r"_(?P<version>vM[0-9]*m[0-9]*p[0-9]*)"
                       r"_(?P<revision>r[0-9]{11})"
                       r"\.bsp$")

CK_REGEX = re.compile(r"^libera_(?P<ck_object>jpss|azrot|elscan)"
                      r"_(?P<utc_start>[0-9]{8}t[0-9]{6})"
                      r"_(?P<utc_end>[0-9]{8}t[0-9]{6})"
                      r"_(?P<version>vM[0-9]*m[0-9]*p[0-9]*)"
                      r"_(?P<revision>r[0-9]{11})"
                      r"\.bc$")

# L0 filename format determined by EDOS Production Data Set and Construction Record filenaming conventions
LIBERA_L0_REGEX = re.compile(r"^(?P<id_char>P|X)"
                             r"(?P<scid>[0-9]{3})"
                             r"(?P<first_apid>[0-9]{4})"
                             # In some cases at least, the last character of the fill field specifies a time (T)
                             # or session (S) based product. e.g. VIIRSSCIENCEAT
                             r"(?P<fill>.{14})"
                             r"(?P<created_time>[0-9]{11})"
                             r"(?P<numeric_id>[0-9])"
                             r"(?P<file_number>[0-9]{2})"
                             r".(?P<extension>PDR|PDS)"
                             r"(?P<signal>.XFR)?$")

LIBERA_L1B_REGEX = re.compile(r"^libera_l1b"
                              r"_(?P<instrument>cam|rad)"
                              r"_(?P<utc_start>[0-9]{8}t[0-9]{6})"
                              r"_(?P<utc_end>[0-9]{8}t[0-9]{6})"
                              r"_(?P<version>vM[0-9]*m[0-9]*p[0-9]*)"
                              r"_(?P<revision>r[0-9]{11})"
                              r"\.(?P<extension>nc|h5)$")

LIBERA_L2_REGEX = re.compile(r"^libera_l2"
                             r"_(?P<product_name>[^_]*)"
                             r"_(?P<utc_start>[0-9]{8}t[0-9]{6})"
                             r"_(?P<utc_end>[0-9]{8}t[0-9]{6})"
                             r"_(?P<version>vM[0-9]*m[0-9]*p[0-9]*)"
                             r"_(?P<revision>r[0-9]{11})"
                             r"\.(?P<extension>nc|h5)$")

MANIFEST_FILE_REGEX = re.compile(r"^libera"
                                 r"_(?P<manifest_type>input|output)"
                                 r"_manifest"
                                 r"_(?P<created_time>[0-9]{8}t[0-9]{6})"
                                 r"\.json")


class DataLevel(Enum):
    """Data product level"""
    L0 = "l0"
    L1B = 'l1b'
    L2 = 'l2'


class ManifestType(Enum):
    """Enumerated legal manifest type values"""
    INPUT = 'INPUT'
    input = INPUT
    OUTPUT = 'OUTPUT'
    output = OUTPUT


class AbstractValidFilename(ABC):
    """Composition of a CloudPath/Path instance with some methods to perform
    regex validation on filenames
    """
    _regex: re.Pattern
    _fmt: str
    _required_parts: tuple

    def __init__(self, *args, **kwargs):
        self.path = AnyPath(*args, **kwargs)

    def __str__(self):
        return str(self.path)

    def __eq__(self, other):
        if self.path == other.path and self.filename_parts == other.filename_parts:
            return True
        return False

    @property
    def path(self):
        """Property containing the file path"""
        return self._path

    @path.setter
    def path(self, new_path: str or Path or CloudPath):
        if isinstance(new_path, str):
            new_path = AnyPath(new_path)
        self.regex_match(new_path)  # validates against regex pattern
        self._path: CloudPath or Path = AnyPath(new_path)

    @property
    def filename_parts(self):
        """Property that contains a namespace of filename parts"""
        return self._parse_filename_parts()

    @classmethod
    def _check_required_parts(cls, local_vars: dict):
        """Checks for the presence of required filename parts

        Parameters
        ----------
        local_vars : dict
            Dictionary of variables passed, created by a call to locals()
        """
        print(local_vars)
        missing = [req for req in cls._required_parts if local_vars[req] is None]
        if missing:
            raise ValueError(
                f"Missing required keyword argument(s) to {cls.__name__}.from_filename_parts: {', '.join(missing)}.")

    @classmethod
    @abstractmethod
    def from_filename_parts(cls,
                            *args,
                            basepath: str or Path = None):
        """Abstract method that must be implemented to provide hinting for required parts"""
        raise NotImplementedError()

    @classmethod
    def _from_filename_parts(cls,
                             basepath: str or Path = None,
                             **parts):
        """Create instance from filename parts.

        The part arg names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        parts :
            Passed directly to _format_filename_parts

        Returns
        -------
        : cls
        """
        filename = cls._format_filename_parts(**parts)
        if basepath is not None:
            return cls(os.path.join(basepath, filename))
        return cls(filename)

    @classmethod
    @abstractmethod
    def _format_filename_parts(cls, **parts):
        """Format parts into a filename

        Note: When this is implemented by concrete classes, **parts becomes a set of explicitly named arguments
        """
        raise NotImplementedError()

    @abstractmethod
    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        # Do stuff to parse the elements of d into a SimpleNamespace
        raise NotImplementedError()

    def regex_match(self, path: str or Path or CloudPath):
        """Parse and validate a given path against class-attribute defined regex

        Returns
        -------
        : dict
        """
        # AnyPath is polymorphic but self.path will always be a CloudPath or Path object with a name attribute.
        match = self._regex.match(path.name)  # pylint: disable=no-member
        if not match:
            raise ValueError(f"Proposed path {path} failed validation against regex pattern {self._regex}")
        return match.groupdict()


class L0Filename(AbstractValidFilename):
    """Filename validation class for L0 files from EDOS."""

    _regex = LIBERA_L0_REGEX
    _fmt = "{id_char}{scid:03}{first_apid:04}{fill:A<14}{created_time}{numeric_id}{file_number:02}.{extension}{signal}"
    _required_parts = (
        'id_char', 'scid', 'first_apid', 'fill', 'created_time', 'numeric_id', 'file_number', 'extension')

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            id_char: str = None,
                            scid: int = None,
                            first_apid: int = None,
                            fill: str = None,
                            created_time: datetime = None,
                            numeric_id: int = None,
                            file_number: int = None,
                            extension: str = None,
                            signal: str = None):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        id_char : str
            Either P (for PDS files, Construction Records) or X (for Delivery Records)
        scid : int
            Spacecraft ID
        first_apid : int
            First APID in the file
        fill : str
            Custom string up to 14 characters long
        created_time : datetime
            Creation time of the file
        numeric_id : int
            Data set ID, 0-9, one digit
        file_number : str
            File number within the data set. Construction records are always file number zero.
        extension : str
            File name extension. Either PDR or PDS
        signal : str or None, Optional
            Optional signal suffix. Always '.XFR'

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        id_char=id_char,
                                        scid=scid,
                                        first_apid=first_apid,
                                        fill=fill,
                                        created_time=created_time,
                                        numeric_id=numeric_id,
                                        file_number=file_number,
                                        extension=extension,
                                        signal=signal)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               id_char: str,
                               scid: int,
                               first_apid: int,
                               fill: str,
                               created_time: datetime,
                               numeric_id: int,
                               file_number: int,
                               extension: str,
                               signal: str = None):
        """Construct a path from filename parts

        Parameters
        ----------
        id_char : str
            Either P (for PDS files, Construction Records) or X (for Delivery Records)
        scid : int
            Spacecraft ID
        first_apid : int
            First APID in the file
        fill : str
            Custom string up to 14 characters long
        created_time : datetime
            Creation time of the file
        numeric_id : int
            Data set ID, 0-9, one digit
        file_number : str
            File number within the data set. Construction records are always file number zero.
        extension : str
            File name extension. Either PDR or PDS
        signal : str or None, Optional
            Optional signal suffix. Always '.XFR'

        Returns
        -------
        : str
            Formatted filename
        """
        signal = signal if signal else ""

        return cls._fmt.format(id_char=id_char,
                               scid=scid,
                               first_apid=first_apid,
                               fill=fill,
                               created_time=created_time.strftime(NUMERIC_DOY_TS_FORMAT),
                               numeric_id=numeric_id,
                               file_number=file_number,
                               extension=extension,
                               signal=signal)

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['scid'] = int(d['scid'])
        d['first_apid'] = int(d['first_apid'])
        d['numeric_id'] = int(d['numeric_id'])
        d['file_number'] = int(d['file_number'])
        d['created_time'] = datetime.strptime(d['created_time'], NUMERIC_DOY_TS_FORMAT)
        return SimpleNamespace(**d)


class L1bFilename(AbstractValidFilename):
    """Filename validation class for L1b products"""

    _regex = LIBERA_L1B_REGEX
    _fmt = "libera_l1b_{instrument}_{utc_start}_{utc_end}_{version}_{revision}.{extension}"
    _required_parts = ('instrument', 'utc_start', 'utc_end', 'version', 'revision', 'extension')

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            instrument: str = None,
                            utc_start: datetime = None,
                            utc_end: datetime = None,
                            version: str = None,
                            revision: datetime = None,
                            extension: str = 'nc'):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        instrument : str
            L2 product type. e.g. cloud-fraction. May contain anything except for underscores.
        utc_start : datetime
            First timestamp in the SPK
        utc_end : datetime
            Last timestamp in the SPK
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime
            Time when the file was created.
        extension : str
            File extension (.nc or .h5)

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        instrument=instrument,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        version=version,
                                        revision=revision,
                                        extension=extension)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               instrument: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               version: str,
                               revision: datetime,
                               extension: str):
        """Construct a path from filename parts

        Parameters
        ----------
        instrument : str
            Libera instrument, cam or rad
        utc_start : datetime
            First timestamp in the SPK
        utc_end : datetime
            Last timestamp in the SPK
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime
            Time when the file was created.
        extension : str
            File extension (.nc or .h5)

        Returns
        -------
        : str
            Formatted filename
        """
        return cls._fmt.format(instrument=instrument,
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               version=version,
                               revision=revision.strftime(REVISION_TS_FORMAT),
                               extension=extension)

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


class L2Filename(AbstractValidFilename):
    """Filename validation class for L2 data products."""
    _regex = LIBERA_L2_REGEX
    _fmt = "libera_l2_{product_name}_{utc_start}_{utc_end}_{version}_{revision}.{extension}"
    _required_parts = ('product_name', 'utc_start', 'utc_end', 'version', 'revision', 'extension')

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            product_name: str = None,
                            utc_start: datetime = None,
                            utc_end: datetime = None,
                            version: str = None,
                            revision: datetime = None,
                            extension: str = 'nc'):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        product_name : str
            L2 product type. e.g. cloud-fraction. May contain anything except for underscores.
        utc_start : datetime
            First timestamp in the SPK
        utc_end : datetime
            Last timestamp in the SPK
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime
            Time when the file was created.
        extension : str
            File extension (.nc or .h5)

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        product_name=product_name,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        version=version,
                                        revision=revision,
                                        extension=extension)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               product_name: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               version: str,
                               revision: datetime,
                               extension: str):
        """Construct a path from filename parts

        Parameters
        ----------
        product_name : str
            L2 product type. e.g. cloud-fraction. May contain anything except for underscores.
        utc_start : datetime
            First timestamp in the SPK
        utc_end : datetime
            Last timestamp in the SPK
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime
            Time when the file was created.
        extension : str
            File extension (.nc or .h5)

        Returns
        -------
        : str
            Formatted filename
        """
        return cls._fmt.format(product_name=product_name,
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               version=version,
                               revision=revision.strftime(REVISION_TS_FORMAT),
                               extension=extension)

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


class ManifestFilename(AbstractValidFilename):
    """Class for naming manifest files"""

    _regex = MANIFEST_FILE_REGEX
    _fmt = "libera_{manifest_type}_manifest_{created_time}.json"
    _required_parts = ('manifest_type', 'created_time')

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            manifest_type: ManifestType = None,
                            created_time: datetime = None):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        manifest_type : ManifestType
            Input or output
        created_time : datetime
            Time of manifest creation (writing).

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        manifest_type=manifest_type,
                                        created_time=created_time)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               manifest_type: ManifestType,
                               created_time: datetime):
        """Construct a path from filename parts

        Parameters
        ----------
        manifest_type : ManifestType
            Input or output
        created_time : datetime
            Time of manifest creation (writing).

        Returns
        -------
        : str
            Formatted filename
        """
        return cls._fmt.format(manifest_type=manifest_type.value.lower(),
                               created_time=created_time.strftime(PRINTABLE_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['manifest_type'] = ManifestType(d['manifest_type'].upper())
        d['created_time'] = datetime.strptime(d['created_time'], PRINTABLE_TS_FORMAT)
        return SimpleNamespace(**d)


class EphemerisKernelFilename(AbstractValidFilename):
    """Class to construct, store, and manipulate an SPK filename"""

    _regex = SPK_REGEX
    _fmt = "libera_{spk_object}_{utc_start}_{utc_end}_{version}_{revision}.bsp"
    _required_parts = ('spk_object', 'utc_start', 'utc_end', 'version', 'revision')

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            spk_object: str = None,
                            utc_start: datetime = None,
                            utc_end: datetime = None,
                            version: str = None,
                            revision: datetime = None):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part arg names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        spk_object : str
            Name of object whose attitude is represented in this SPK.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime
            When the file was last revised.

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        spk_object=spk_object,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        version=version,
                                        revision=revision)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               spk_object: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               version: str,
                               revision: datetime):
        """Create an instance from a given path

        Parameters
        ----------
        spk_object : str
            Name of object whose ephemeris is represented in this SPK.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime
            Time when the file was last revised

        Returns
        -------
        : cls
        """
        return cls._fmt.format(spk_object=spk_object,
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               version=version,
                               revision=revision.strftime(REVISION_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


class AttitudeKernelFilename(AbstractValidFilename):
    """Class to construct, store, and manipulate an SPK filename"""

    _regex = CK_REGEX
    _fmt = "libera_{ck_object}_{utc_start}_{utc_end}_{version}_{revision}.bc"
    _required_parts = ('ck_object', 'utc_start', 'utc_end', 'version', 'revision')

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            ck_object: str = None,
                            utc_start: datetime = None,
                            utc_end: datetime = None,
                            version: str = None,
                            revision: datetime = None):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part arg names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        ck_object : str
            Name of object whose attitude is represented in this CK.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime
            When the file was last revised.

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        ck_object=ck_object,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        version=version,
                                        revision=revision)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               ck_object: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               version: str,
                               revision: datetime):
        """Create an instance from a given path

        Parameters
        ----------
        ck_object : str
            Name of object whose attitude is represented in this CK.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime
            When the file was last revised.

        Returns
        -------
        : cls
        """
        return cls._fmt.format(ck_object=ck_object,
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               version=version,
                               revision=revision.strftime(REVISION_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


def get_current_revision_str():
    """Get the current `r%y%j%H%M%S` string for filename revisions.

    Returns
    -------
    : str
        Current (now) revision string.
    """
    return datetime.utcnow().strftime(REVISION_TS_FORMAT)


def format_semantic_version(semantic_version: str):
    """Formats a semantic version string X.Y.Z into a filename-compatible string like vMXmYpZ, for Major, minor, patch.

    Parameters
    ----------
    semantic_version : str
        String matching X.Y.Z where X, Y and Z are integers of any length

    Returns
    -------
    : str
    """
    major, minor, patch = semantic_version.split('.')
    return f"vM{major}m{minor}p{patch}"


def get_current_version_str(package_name: str):
    """Retrieve the current version of a (algorithm) package and format it for inclusion in a filename

    Parameters
    ----------
    package_name : str
        Package for which to retrieve a version string. This should be your algorithm package and it must use a
        semantic versioning scheme, configured in project metadata.

    Returns
    -------
    : str
        Version string in format vM1m2p3
    """
    semver = metadata.version(package_name)
    return format_semantic_version(semver)



"""
Module containing Mixin classes for providing common functionality to ORM objects.
"""
# Installed
from sqlalchemy import inspect
from sqlalchemy.orm import Session, Query


class ReprMixin:
    """
    Mixin class to provide a nice and configurable __repr__ method. To control which attributes
    get included in __repr__, use __repr_attrs__ on the class mixing this in.

    Yanked from https://github.com/absent1706/sqlalchemy-mixins
    """
    __abstract__ = True
    __repr_attrs__ = []
    __repr_max_length__ = 36

    def __init__(self, *args, **kwargs):
        # TODO: Remove this after PyCharm updates past 2021.3.2. It's only here to suppress an IDE warning.
        pass

    @property
    def _id_str(self):
        """
        Figure out what the ID string should be.

        Returns
        -------
        : str
        """
        ids = inspect(self).identity
        if ids:
            return '-'.join([str(x) for x in ids]) if len(ids) > 1 \
                   else str(ids[0])
        return 'None'

    @property
    def _repr_attrs_str(self):
        """
        Create the string representation of the attributes listed in __repr_attrs__

        Returns
        -------
        : str
        """
        max_length = self.__repr_max_length__

        values = []
        single = len(self.__repr_attrs__) == 1
        for key in self.__repr_attrs__:
            if not hasattr(self, key):
                raise KeyError(f"{self.__class__} has incorrect attribute '{key}' in __repr__attrs__")
            value = getattr(self, key)
            wrap_in_quote = isinstance(value, str)

            value = str(value)
            if len(value) > max_length:
                value = value[:max_length] + '...'

            if wrap_in_quote:
                value = f"'{value}'"
            values.append(value if single else f"{key}:{value}")

        return ' '.join(values)

    def __repr__(self):
        # Get id like '#123'
        id_str = ('#' + self._id_str) if self._id_str else ''
        # Join class name, id and repr_attrs
        return f"<{self.__class__.__name__} {id_str}{f' {self._repr_attrs_str}' if self._repr_attrs_str else ''}>"


class DataProductMixin:
    """
    Mixin class that provides methods that are commonly used in data product ORM classes.
    """
    __abstract__ = True

    def __init__(self, *args, **kwargs):
        # TODO: Remove this after PyCharm updates past 2021.3.2. It's only here to suppress an IDE warning.
        pass

    @classmethod
    def _filter(cls, query: Query, **filters):
        """ Filters an existing query object for a set of keyword filters """
        for attr, value in filters.items():
            query.filter(getattr(cls, attr) == value)
        return query

    @classmethod
    def latest(cls, session: Session = None, **filters):
        """
        Finds the latest products (highest revision), filtered by **filters**

        Parameters
        ----------
        session: Session, Optional
            If None, creates a new session
        filters: dict
            Optional filters to narrow down results

        Returns
        -------
        : list
        """
        # TODO: Implement logic for finding the latest revision of each "unique observation", whatever that ends up
        #    meaning. For BDS products, that could be a specific timerange or just a date.
        pass

    @classmethod
    def flagged(cls, session: Session = None, **filters):
        """
        Queries products with quality flags, optionally filtered by **filters**

        Parameters
        ----------
        session: Session, Optional
            If None, creates a new session
        filters: dict
            Optional filters to narrow results

        Returns
        -------
        : list
        """
        # TODO: Implement retrieval of quality-flagged data products once quality flags are defined
        pass



"""ORM objects for SQLAlchemy"""
# Installed
from sqlalchemy import Column, FetchedValue, ForeignKey
from sqlalchemy.dialects.postgresql import SMALLINT, BIGINT, INTEGER, TEXT, TIMESTAMP, BOOLEAN, NUMERIC
from sqlalchemy.orm import relationship
# Local
from libera_utils.db import Base
from libera_utils.db.mixins import ReprMixin, DataProductMixin


class Cr(Base, ReprMixin):
    """
    Construction record
    """
    __repr_attrs__ = ["filename"]

    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    file_name = Column(TEXT)
    ingested = Column(TIMESTAMP(timezone=True), server_default=FetchedValue())
    archived = Column(TIMESTAMP(timezone=True))
    edos_software_version = Column(INTEGER)
    construction_record_type = Column(INTEGER)
    test_flag = Column(BOOLEAN)
    n_scs_start_stops = Column(INTEGER)
    n_bytes_fill_data = Column(NUMERIC(20))
    n_length_mismatches = Column(BIGINT)
    first_packet_sc_time = Column(NUMERIC(20))
    last_packet_sc_time = Column(NUMERIC(20))
    first_packet_utc_time = Column(TIMESTAMP(timezone=True))
    last_packet_utc_time = Column(TIMESTAMP(timezone=True))
    first_packet_esh_time = Column(NUMERIC(20))
    last_packet_esh_time = Column(NUMERIC(20))
    n_rs_corrections = Column(BIGINT)
    n_packets = Column(BIGINT)
    size_bytes = Column(NUMERIC(20))
    n_ssc_discontinuities = Column(INTEGER)
    completion_time = Column(NUMERIC(20))
    n_apids = Column(SMALLINT)
    n_pds_files = Column(SMALLINT)

    scs_start_stop_times: list = relationship("CrScsStartStopTimes", back_populates="construction_record")
    apids: list = relationship("CrApid", back_populates="construction_record")
    pds_files: list = relationship("PdsFile", back_populates="construction_record")


class CrApid(Base, ReprMixin):  # I know...
    """
    APID data record in a CR
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    cr_id = Column(INTEGER, ForeignKey(Cr.id))
    scid_apid = Column(BIGINT)
    byte_offset = Column(NUMERIC(20))
    n_vcids = Column(SMALLINT)
    n_ssc_gaps = Column(BIGINT)
    n_edos_generated_fill_data = Column(BIGINT)
    count_edos_generated_octets = Column(NUMERIC(20))
    n_length_discrepancy_packets = Column(BIGINT)
    first_packet_sc_time = Column(NUMERIC(20))
    last_packet_sc_time = Column(NUMERIC(20))
    first_packet_utc_time = Column(TIMESTAMP(timezone=True))
    last_packet_utc_time = Column(TIMESTAMP(timezone=True))
    esh_first_packet_time = Column(NUMERIC(20))
    esh_last_packet_time = Column(NUMERIC(20))
    n_vcdu_corrected_packets = Column(BIGINT)
    n_in_the_data_set = Column(BIGINT)
    n_octect_in_apid = Column(NUMERIC(20))

    construction_record = relationship("Cr", back_populates="apids")
    vcids: list = relationship("CrApidVcid", back_populates="apid")
    ssc_gaps: list = relationship("CrApidSscGap", back_populates="apid")
    ssc_length_discrepancies: list = relationship("CrApidSscLenDiscrepancies", back_populates="apid")
    edos_fill_data: list = relationship("CrApidEdosGeneratedFillData", back_populates="apid")


class CrApidVcid(Base, ReprMixin):
    """
    VCID included in PDS for a single APID
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    cr_apid_id = Column(INTEGER, ForeignKey(CrApid.id))
    scid_vcid = Column(INTEGER)

    apid = relationship("CrApid", back_populates="vcids")


class CrApidSscGap(Base, ReprMixin):
    """
    SSC gap described per APID in a CR
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    cr_apid_id = Column(INTEGER, ForeignKey(CrApid.id))
    first_missing_ssc = Column(BIGINT)
    gap_byte_offset = Column(NUMERIC(20))
    n_missing_sscs = Column(BIGINT)
    preceding_packet_sc_time = Column(NUMERIC(20))
    following_packet_sc_time = Column(NUMERIC(20))
    preceding_packet_utc_time = Column(TIMESTAMP(timezone=True))
    following_packet_utc_time = Column(TIMESTAMP(timezone=True))
    preceding_packet_esh_time = Column(NUMERIC(20))
    following_packet_esh_time = Column(NUMERIC(20))

    apid = relationship("CrApid", back_populates="ssc_gaps")


class PdsFile(Base, ReprMixin):
    """
    Record of a single file as part of a PDS
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    cr_id = Column(INTEGER, ForeignKey(Cr.id))
    file_name = Column(TEXT)
    ingested = Column(TIMESTAMP(timezone=True))
    archived = Column(TIMESTAMP(timezone=True))

    construction_record = relationship("Cr", back_populates="pds_files")
    apids: list = relationship("PdsFileApid", back_populates="pds_file")


class PdsFileApid(Base, ReprMixin):
    """
    APID information for a single PDS file
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    pds_file_id = Column(INTEGER, ForeignKey(PdsFile.id))
    scid_apid = Column(BIGINT)
    first_packet_sc_time = Column(NUMERIC(20))
    last_packet_sc_time = Column(NUMERIC(20))
    first_packet_utc_time = Column(TIMESTAMP(timezone=True))
    last_packet_utc_time = Column(TIMESTAMP(timezone=True))

    pds_file = relationship("PdsFile", back_populates="apids")


class CrScsStartStopTimes(Base, ReprMixin):
    """
    SSC Start and Stop Time information for a Construction Record
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    cr_id = Column(INTEGER, ForeignKey(Cr.id))
    scs_start_sc_time = Column(NUMERIC(20))
    scs_stop_sc_time = Column(NUMERIC(20))
    scs_start_utc_time = Column(TIMESTAMP(timezone=True))
    scs_stop_utc_time = Column(TIMESTAMP(timezone=True))

    construction_record = relationship("Cr", back_populates="scs_start_stop_times")


class CrApidSscLenDiscrepancies(Base, ReprMixin):
    """
    SSC length discrepancies per APID in a CR
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    cr_apid_id = Column(INTEGER, ForeignKey(CrApid.id))
    ssc_length_discrepancy = Column(BIGINT)

    apid = relationship("CrApid", back_populates="ssc_length_discrepancies")


class CrApidEdosGeneratedFillData(Base, ReprMixin):
    """
    EDOS generated fill data per APID in a CR
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    cr_apid_id = Column(INTEGER, ForeignKey(CrApid.id))
    ssc_with_generated_data = Column(BIGINT)
    filled_byte_offset = Column(NUMERIC(20))
    index_to_fill_octet = Column(BIGINT)

    apid = relationship("CrApid", back_populates="edos_fill_data")


class SpkCkFile(Base, ReprMixin, DataProductMixin):
    """
    SPICE CK and SPK kernel file
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    file_name = Column(TEXT)
    start_sc_time = Column(NUMERIC(20))
    stop_sc_time = Column(NUMERIC(20))
    start_utc_time = Column(TIMESTAMP(timezone=True))
    stop_utc_time = Column(TIMESTAMP(timezone=True))
    revision = Column(INTEGER)
    quality_flag = Column(INTEGER)

    pds_files: list = relationship("PdsFile", secondary='sdp.spk_ck_file_pds_file_jt')


class SpkCkFilePdsFileJt(Base):
    """
    Join table between PdsFile and SpkCkFile for n:m relationship
    """
    pds_file_id = Column(INTEGER, ForeignKey(PdsFile.id), primary_key=True)
    spk_ck_file_id = Column(INTEGER, ForeignKey(SpkCkFile.id), primary_key=True)


class L1bCam(Base):
    """
    L1b Camera product files
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    file_name = Column(TEXT)
    revision = Column(INTEGER)
    quality_flag = Column(INTEGER)


class L1bRad(Base):
    """
    L1b Radiometer product files
    """
    id = Column(INTEGER, primary_key=True, server_default=FetchedValue())
    file_name = Column(TEXT)
    revision = Column(INTEGER)
    quality_flag = Column(INTEGER)



"""Module for manifest file handling"""
import warnings
# Standard
from datetime import datetime
import json
import logging
from pathlib import Path
from hashlib import md5
# Installed
from cloudpathlib import S3Path, AnyPath
# Local
from libera_utils.io.smart_open import smart_open
from libera_utils.io.filenaming import ManifestFilename, ManifestType

logger = logging.getLogger(__name__)


class ManifestError(Exception):
    """Generic exception related to manifest file handling"""
    pass


class Manifest:
    """Object representation of a JSON manifest file"""

    __manifest_elements = (
        "manifest_type",
        "files",
        "configuration"
    )

    def __init__(self, manifest_type: ManifestType,
                 files: list = None, configuration: dict = None, filename: str or ManifestFilename = None):
        """Constructor

        Parameters
        ----------
        manifest_type : ManifestType
            Type of manifest
        files : list, Optional
            List of dictionaries. Each entry must contain a `filename` key and a `checksum` key.
        configuration : dict, Optional
            Freeform dictionary of configuration items. It's up to the consumer to understand this JSON object.
        filename : str or ManifestFilename, Optional
            Preset filename. Must be a ManifestFilename object or a str representing a valid manifest file path.
        """
        self.manifest_type = manifest_type if isinstance(manifest_type, ManifestType) else ManifestType(manifest_type)
        self.files = files if files else []
        self.configuration = configuration if configuration else {}
        if filename:
            self.filename = filename if isinstance(filename, ManifestFilename) else ManifestFilename(filename)
        else:
            self.filename = None

    def __str__(self):
        return f"<Manifest:{self.filename if self.filename else '(unnamed)'} " \
               f"{self.manifest_type} of {len(self.files)} files>"

    def _generate_filename(self):
        """Generate a valid manifest filename"""
        mfn = ManifestFilename.from_filename_parts(
            manifest_type=self.manifest_type,
            created_time=datetime.utcnow()
        )
        return mfn

    def to_json_dict(self):
        """Create a dict representation suitable for writing out.

        Returns
        -------
        : dict
        """
        valid_json = json.dumps({
            'manifest_type': self.manifest_type.value,
            'files': self.files,
            'configuration': self.configuration
        })
        return json.loads(valid_json)

    @classmethod
    def from_file(cls, filepath: str or Path or S3Path):
        """Read a manifest file and return a Manifest object (factory method).

        Parameters
        ----------
        filepath : str or Path or S3Path
            Location of manifest file to read.

        Returns
        -------
        : Manifest
        """
        with smart_open(filepath) as manifest_file:
            contents = json.loads(manifest_file.read())
        for element in cls.__manifest_elements:
            if element not in contents:
                raise ManifestError(f"{filepath} is not a valid manifest file. Missing required element {element}.")
        obj = cls(ManifestType(contents['manifest_type'].upper()),
                  contents['files'],
                  contents['configuration'],
                  filename=filepath)
        obj.validate()  # If we just read it, it should be valid
        return obj

    def write(self, outpath: str or Path or S3Path, filename: str = None):
        """Write a manifest file from a Manifest object (self).

        Parameters
        ----------
        outpath : str or Path or S3Path
            Directory path to write to (directory being used loosely to refer also to an S3 bucket path).
        filename : str, Optional
            Optional filename, must be a valid manifest filename.
            If not provided, the method uses the objects internal filename attribute. If that is
            not set, then a filename is automatically generated.

        Returns
        -------
        : Path or S3Path
        """
        if filename is None:
            if self.filename is None:
                filename = self._generate_filename()
            else:
                filename = self.filename
        else:
            filename = ManifestFilename(filename)

        filepath = AnyPath(outpath) / filename.path
        self.filename = ManifestFilename(filepath)  # Update object's filename to the filepath we just wrote
        self.validate()  # Final check before writing
        with smart_open(self.filename.path, 'x') as manifest_file:
            json.dump(self.to_json_dict(), manifest_file)
        return self.filename.path

    def validate(self):
        """Validate the contents of this manifest object"""
        if not isinstance(self.files, list):
            raise ValueError("The files attribute must be a dictionary.")
        if not isinstance(self.configuration, dict):
            raise ValueError("The configuration attribute must be a dictionary.")
        if not isinstance(self.manifest_type, ManifestType):
            raise ValueError("The manifest_type attribute must be a ManifestType object.")
        if not isinstance(self.filename, ManifestFilename):
            raise ValueError("The filename attribute must be a ManifestFilename object.")
        for filedict in self.files:
            if tuple(filedict.keys()) != ('filename', 'checksum'):
                raise ValueError("Each entry of the files attribute must be a dictionary containing the keys "
                                 f"`filename` and `checksum`. Got {tuple(filedict.keys())}.")

    def validate_checksums(self):
        """Validate checksums of listed files"""
        failed_filenames = []
        for record in self.files:
            checksum_expected = record['checksum']
            filename = record['filename']
            # Validate checksums
            with smart_open(filename, 'rb') as fh:
                checksum_calculated = md5(fh.read(), usedforsecurity=False).hexdigest()
                if checksum_expected != checksum_calculated:
                    logger.error(f"Checksum validation for {filename} failed. "
                                 f"Expected {checksum_expected} but got {checksum_calculated}.")
                    failed_filenames.append(str(filename))
        if failed_filenames:
            raise ValueError(f"Files failed checksum validation: {', '.join(failed_filenames)}")

    def add_files(self, *files):
        """Add a file to the manifest from filename
        Parameters
        ----------
        files : str or Path or S3Path
            Path to the file to add to the manifest.

        Returns
        -------
        None
        """

        for file in files:
            if AnyPath(file).name in (AnyPath(fs['filename']).name for fs in self.files):
                warnings.warn(f"Attempting to add {file} to manifest {self} but it is already included.")
                continue
            with smart_open(file) as fh:
                checksum_calculated = md5(fh.read(), usedforsecurity=False).hexdigest()
            if checksum_calculated in (fs['checksum'] for fs in self.files):
                warnings.warn(f"Attempting to add {file} to manifest {self} but another file with "
                              f"the same checksum is already included.")
            file_structure = {"filename": str(file),
                              "checksum": str(checksum_calculated)}
            self.files.append(file_structure)

    def add_file_to_manifest(self, file):
        """Deprecated legacy method replaced by add_files"""
        warnings.warn("add_file_to_manifest(file) is deprecated. Use add_files(*files) instead",
                      category=DeprecationWarning)
        self.add_files(file)

    def add_desired_time_range(self, start_datetime: datetime, end_datetime: datetime):
        """Add a file to the manifest from filename
          Parameters
        ----------
        start_datetime : datetime
            The desired start time for the range of data in this manifest

        end_datetime : datetime
            The desired end time for the range of data in this manifest

        Returns
        -------
        None
        """
        self.configuration["start_time"] = start_datetime.strftime('%Y-%m-%d:%H:%M:%S')
        self.configuration["end_time"] = end_datetime.strftime('%Y-%m-%d:%H:%M:%S')
        


"""Module for packet ingest"""
# Standard
import argparse
import datetime
import logging
import os
# Installed
from cloudpathlib import AnyPath
from sqlalchemy import func
# Local
from libera_utils.db import getdb
from libera_utils.io.construction_record import ConstructionRecord, PDSRecord
from libera_utils.io.manifest import Manifest, ManifestType, ManifestFilename
from libera_utils.db.models import Cr, PdsFile
from libera_utils.io.smart_open import smart_copy_file


logger = logging.getLogger(__name__)


def ingest(parsed_args: argparse.Namespace):
    """Ingest and update records into database using manifest
    Parameters
    ----------
    parsed_args : argparse.Namespace
        Namespace of parsed CLI arguments

    Returns
    -------
    output_manifest_path : str
        Path of output manifest
    """

    processing_dropbox = os.environ['PROCESSING_DROPBOX']

    # read json information
    m = Manifest.from_file(parsed_args.manifest_filepath)
    m.validate_checksums()

    mfn = ManifestFilename.from_filename_parts(
        manifest_type=ManifestType.OUTPUT,
        created_time=datetime.datetime.utcnow())

    db_pds_dict_all = {}
    output_files = []

    for file in m.files:
        # is there a next cr in the manifest
        if 'CONS' in file['filename']:
            db_pds_dict, con_ingested_dict = cr_ingest(file, processing_dropbox)
            db_pds_dict_all.update(db_pds_dict)
            output_files.append(con_ingested_dict)

    for file in m.files:
        # is there a next pds in the manifest
        if 'PDS' in file['filename']:
            pds_ingested_dict = pds_ingest(file, processing_dropbox)
            if pds_ingested_dict:
                output_files.append(pds_ingested_dict)

    # insert cr_id for pds files in the db associated with the current cr
    if db_pds_dict_all:
        with getdb().session() as s:
            for cr_filename in db_pds_dict_all.items():
                # query cr_id that has been inserted
                cr_query = s.query(Cr).filter(Cr.file_name == cr_filename[0]).all()
                # query all pds associated with cr
                pds_query = s.query(PdsFile).filter(
                    PdsFile.file_name == func.any(cr_filename[1])).all()
                # assign cr_id
                for pds in pds_query:
                    pds.cr_id = cr_query[0].id

    # write output manifest to L0 ingest dropbox
    output_dir = AnyPath(processing_dropbox)
    logger.info("Writing resulting output manifest to %s", output_dir)

    # Write output manifest file containing a list of the product files that the processing created
    output_manifest_path = os.path.join(output_dir, str(mfn))
    output_manifest = Manifest(manifest_type=ManifestType.OUTPUT,
                               filename=output_manifest_path,
                               files=output_files,
                               configuration={})

    # move files over
    for file in output_files:
        input_dir = os.path.join(os.path.dirname(m.files[0]['filename']),
                                 os.path.basename(file['filename']))
        smart_copy_file(input_dir, output_dir, delete=parsed_args.delete)

    output_manifest.write(output_manifest_path)
    logger.info("Algorithm complete. Exiting.")

    return output_manifest_path


def cr_ingest(file: dict, output_dir: str):
    """Ingest cr records into database
    Parameters
    ----------
    file : Dictionary
        Dictionary containing path and checksum of cr
    output_dir : str
        Directory for output data

    Returns
    -------
    db_pds_dict : Dictionary
        Dictionary that associates the pds file in the db with the current cr
    ingested_dict : Dictionary
        Dictionary of records that have been ingested
    """
    filename = os.path.basename(file['filename'])
    db_pds = []

    with getdb().session() as s:

        cr_query = s.query(Cr).filter(
            Cr.file_name == filename).all()

        # check if cr is in the db
        if not cr_query:

            # parse cr into nested orm objects
            cr = ConstructionRecord.from_file(file['filename'])

            if not cr.pds_files_list:
                all_pds_dict = {}
            else:
                all_pds_dict = {f.pds_filename: f for f in cr.pds_files_list}

            pds_query = s.query(PdsFile).filter(
                PdsFile.file_name == func.any(list(all_pds_dict.keys()))).all()

            # if there are some pds records from the current cr in the db
            # associate them with current cr, but do not set pds ingest time
            if pds_query:

                for i, pds_object in enumerate(pds_query):
                    db_pds.append(pds_query[i].file_name)
                    logger.info("In database: %s", pds_object.file_name)

                    if pds_query[i].file_name in list(all_pds_dict):
                        cr.pds_files_list.remove(
                            all_pds_dict[pds_query[i].file_name])

            cr_orm = cr.to_orm()
            s.merge(cr_orm)

            # output filename
            output_dir = os.path.join(output_dir, filename)

            # create ingested dictionary
            ingested_dict = {"filename": output_dir,
                             "checksum": file['checksum']}
        else:
            logger.info("Duplicate cr: %s", filename)
            ingested_dict = {}

    # for the pds files that were already in the db,
    # associate the pds file in the db with the current cr
    if db_pds:
        db_pds_dict = {filename: db_pds}
    else:
        db_pds_dict = {}

    return db_pds_dict, ingested_dict


def pds_ingest(file: dict, output_dir: str):
    """Ingest pd records into database that do not have an associated cr
    Parameters
    ----------
    file : Dictionary
        Dictionary containing path and checksum of pd
    output_dir : str
        Directory for output data

    Returns
    -------
    ingested_dict : Dictionary
        Dictionary of records that have been ingested
    """
    filename = os.path.basename(file['filename'])

    with getdb().session() as s:

        # check to see if pds is in db
        pds_query = s.query(PdsFile).filter(
            PdsFile.file_name == filename).all()

        # if pds is not in db then insert the pds file into the db
        # without associating it with a cr; set the ingest time
        if not pds_query:
            # parse pds into nested orm objects
            pds = PDSRecord(filename)
            pds_orm = pds.to_orm()
            s.add(pds_orm)

            # output filename
            output_dir = os.path.join(output_dir, filename)

            # create ingested dictionary
            ingested_dict = {"filename": output_dir,
                             "checksum": file['checksum']}
        # if pds is in db but does not have ingest time, update the ingest time
        elif pds_query[0].ingested is None:
            pds_query[0].ingested = datetime.datetime.utcnow()

            # output filename
            output_dir = os.path.join(output_dir, filename)

            # create ingested dictionary
            ingested_dict = {"filename": output_dir,
                             "checksum": file['checksum']}
        else:
            logger.info("Duplicate pd: %s", filename)
            ingested_dict = {}

    return ingested_dict
